#!/usr/bin/env python3
#
# Plot E field of charge which starts from rest and accelerates to constant velocity
# Matthew Chapman <contact@zmatt.net>
#

import numpy as np
import numpy.linalg as linalg
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.collections import LineCollection
from matplotlib.patches import Circle

v = 0.9  # relative to c

# initial acceleration during t=0..taccel and x=0..xaccel
a = 10
taccel = v/a
xaccel = a*taccel*taccel/2

# plot ranges
xymax = 1
tmax = 1
ntstep = 20
ts = np.linspace(0, tmax, ntstep)

def sourcepos(t):
  return xaccel+v*(t-taccel) if t >= taccel else a*t*t/2

# determine source of current observation assuming source at x=v*t
def wavesource_constv(x, y, t):
  # quadratic, discriminant always >=0 for v<=1
  discriminant = (v*t-x)**2 + (1-v*v)*y*y
  tr = (-(v*x-t) - np.sqrt(discriminant)) / (1-v*v)
  return tr

# determine source of current observation assuming source at x=a*t*t/2
def wavesource_accel(x, y, t):
  # this code solves a depressed quartic
  A = -4*(1+a*x)/(a*a)
  B = 8*t/(a*a)
  C = 4*(x*x+y*y-t*t)/(a*a)
  p = -A*A/12 - C
  q = -A*A*A/108 + A*C/3 - B*B/8
  w = (-q/2+np.sqrt(q*q/4+p*p*p/27+0J))**(1./3)
  Y = A/6 + w - p/(3*w)
  Z1 = np.sqrt(2*Y-A)/2
  Z2 = np.sqrt(-2*Y-A+2*B/np.sqrt(2*Y-A))/2
  tr = abs(Z1-Z2)  # smallest positive solution
  assert abs(np.imag(tr)) < 1e-6
  tr = np.real(tr)
  return tr

def wavesource(x, y, t):
  if x*x+y*y >= t*t:
    # observer has not seen charge start moving yet
    X = 0
    V = 0
    A = 0
    tr = 0
  else:
    # first solve assuming constv region, if time negative then still in accelerating region
    tr = wavesource_constv(x-xaccel, y, t-taccel)
    if tr >= 0:
      X = v*tr + xaccel
      V = v
      A = 0
      tr = tr + taccel
    else:
      tr = wavesource_accel(x, y, t)
      assert tr >= 0
      X = a*tr*tr/2
      V = a*tr
      A = a
  return (np.array((X,0,0)), np.array((V,0,0)), np.array((A,0,0)), tr)


def Efield(x, y, t):
  robs = np.array((x, y, 0))
  X, V, A, tr = wavesource(x, y, t)
  rvec = robs - X
  r = linalg.norm(rvec)
  assert(tr == 0 or abs(t - tr - r) < 1e-6)  # sanity check tr calculation
  if r == 0:
    r = 1e-12
  n = rvec/r
  gamma = 1/np.sqrt(1-np.dot(V,V))
  E = (n-V)/(gamma*gamma*r*r) + np.cross(n, np.cross(n-V, A))/r
  E /= (1-np.dot(V, n))**3
  return E[:2]

# trick: we know the field is symmetric at infinity, so we draw outside in
def fieldstartpoints(t):
  nlines = 32
  angles = np.linspace(0, 2*np.pi, nlines, endpoint=False)
  startpoints = []
  for angle in angles:
    # field must be symmetric beyond tmax, but draw at least to axis
    a = angle % (np.pi/2)
    a = np.pi/2-a if (a > np.pi/4) else a
    dist2axis = 1/np.cos(a)
    r = max(tmax, dist2axis)
    startpoints.append((r*np.cos(angle), r*np.sin(angle)))
  return startpoints

# second-order Runge-Kutta, specifically Ralston's method
def rk2(f, x):
  h = 0.01
  k1 = f(x)
  k1 = h * (k1 / linalg.norm(k1))
  est = x + k1*3/4
  k2 = f(est)
  k2 = h * (k2 / linalg.norm(k2))
  return x + k1/3 + k2*2/3

class FieldPlot:
  def __init__(self, lc, c):
    self.lc = lc
    self.c = c

  def remove(self):
    self.lc.remove()
    self.c.remove()

def fieldplot(ax, sourcex, startpoints, f):
  fieldlines = []
  for x, y in startpoints:
    segs = []
    for i in range(1000):
      segs.append((x,y))
      x, y = rk2(f, (x, y))
      rsq = (x-sourcex)**2 + y**2
      if abs(rsq) < 1e-4:
        break
    fieldlines.append(segs)

  lc = ax.add_collection(LineCollection(fieldlines))
  c = ax.add_patch(Circle((sourcex,0), 0.02, color='r', zorder=1000))
  return FieldPlot(lc, c)

def animate(ti):
  print("Time step %d..." % ti)
  t = ts[ti]

  global fieldplotobj
  if 'fieldplotobj' in globals():
    fieldplotobj.remove()

  sourcex = sourcepos(t)
  startpoints = fieldstartpoints(ts[ti])
  fieldplotobj = fieldplot(ax, sourcex, startpoints, lambda x: -Efield(x[0], x[1], ts[ti]))

fig, ax = plt.subplots()
ax.set_aspect('equal')
ax.set_xlim(-xymax, xymax)
ax.set_ylim(-xymax, xymax)
anim = animation.FuncAnimation(fig, animate, frames=ntstep, blit=False, repeat=False, init_func=lambda: None)
#anim.save('./efield-accel.mp4', writer='ffmpeg', fps=5)
plt.show()
